/***************************************************************************
                          FBdraw.c  -  Drawing primitives
                             -------------------
    begin                : Mon Nov 20 2000
    copyright            : (C) 2000 by Daniele Venzano
    email                : venza@iol.it

    Riscitto FB_line() il 16 aprile 2001, R. A. Foglietta <robang@libero.it> 
    Modifiche rilasciate sotto licenza GPL come beta software.
    TODO: ottimizzare la velocit di disegno delle linee.

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License.               *
 *                                                                         *
 ***************************************************************************/
#include "FBlib.h"
#include "FBpriv.h"

int FB_putpixel(int x, int y, FB_pixel color)
{
	long int location;

	location=x * (work_var_info.bits_per_pixel/8) + y * (fb_fix_info.line_length);
	if(!(location > visiblesize))
	{
		if(work_var_info.bits_per_pixel == 16)
		{
			*(fb_visp + location) = (u_char)color;
			*(fb_visp + location + 1) = (u_char)(color>>8);
		}
		else if (work_var_info.bits_per_pixel == 32)
		{
			*(fb_visp + location) = (u_char)color;
			*(fb_visp + location + 1) = (u_char)(color>>8);
			*(fb_visp + location + 2) = (u_char)(color>>16);
			*(fb_visp + location + 3) = (u_char)(color>>24);
		}
		else return -1;
	}
	return 0;
}

FB_pixel FB_getpixel(int x, int y)
{
	long int location;
	FB_pixel tmp;
	
	location=x * (work_var_info.bits_per_pixel/8) + y * (fb_fix_info.line_length);
	if(location > visiblesize)
		return 0; /* return black if out of bounds (weird) */
	tmp = *(fb_visp + location) | (*(fb_visp + location + 1)<<8);
	if(work_var_info.bits_per_pixel == 32)
		tmp |= (*(fb_visp + location + 2)<<16) + (*(fb_visp + location + 3)<<24);
	return tmp;
}

int FB_vline(int x, int sy, int ey, FB_pixel color)
{
	int i,err=0;
	
	if(sy > ey)
	{
		i=sy;
		sy=ey;
		ey=i;
	}
	for(i=sy;i<=ey;i++)
		err+=FB_putpixel(x, i, color);
	return err;
}

int FB_hline(int sx, int ex, int y, FB_pixel color)
{
	int i,err=0;
	
	if(sx > ex)
	{
		i=sx;
		sx=ex;
		ex=i;
	}
	for(i=sx;i<=ex;i++)
		err+=FB_putpixel(i, y, color);
	return err;
}


int FB_line(int sx, int sy, int ex, int ey, FB_pixel color)
{
 /* (C) 2001 Roberto A. Foglietta, <robang@libero.it>
  *          delivered under GPL as beta software
  */ 

 float dx,dy;
 float nx, ny, fx, fy, sty, stx;

 if(sx == ex)
  FB_vline(sx,sy,ey,color);
 else if(sy == ey)
  FB_hline(sx,ex,sy,color);
 else 
 {
  dx=abs(ex-sx);
  dy=abs(ey-sy);

  FB_putpixel(sx, sy, color);
  FB_putpixel(ex, ey, color);

  if(dy < dx) 
  {     
   if(sx > ex) 
   {
    nx = ex+1;
    fx = sx;
    ny = ey;         
    sty = -(float)(ey-sy)/dx;
   }
   else
   {
    nx = sx+1;
    fx = ex;
    ny = sy;         
    sty = (float)(ey-sy)/dx;
   }
   for(; nx < fx; nx++) 
   {
    ny += sty;
    FB_putpixel((int)nx, (int)ny, color);
   }
  }
  else if(dx <= dy)
  {
   if(sy > ey)
   {
    ny = ey+1;
    fy = sy;
    nx = ex;     
    stx = -(float)(ex-sx)/dy;
   }
   else
   {
    ny = sy+1;
    fy = ey;
    nx = sx;     
    stx = (float)(ex-sx)/dy;
   }
   for(; ny < fy; ny++)
   {
    nx += stx;
    FB_putpixel((int)nx, (int)ny, color);
   }
  }
/*        else
         for(ny = sy; ny < ey; ny++)
          FB_putpixel((int)(nx++), (int)ny, color);

  */
 }
 return 0;

}


/* This is Bresenham's line algorithm taken from ofbis
int FB_line(int sx, int sy, int ex, int ey, FB_pixel color)
{
	int dx,dy;
	int x,y,endx,endy;
	int step,i,c1,c2;
	
	if(sx == ex)
		FB_vline(sx,sy,ey,color);
	if(sy == ey)
		FB_hline(sx,ex,sy,color);
		
	dx=abs(ex-sx);
	dy=abs(ey-sy);
	if(dx > dy)
	{
		x= (sx<ex ? sx : ex);
		endx = (sx<ex ? ex : sx);
		if(x == sx)
		{
			y=sy;
			step=((ey-y)<0 ? -1 : 1);
		}
		else
		{
			y=ey;
			step=((sy-y)<0 ? -1 : 1);
		}
		i = 2*dy - dx;
		c1=2*dy;
		c2=2*(dy-dx);
		FB_putpixel(x,y,color);
		while(x++ <= endx)
		{
			if(i<0)
				i += c1;
			else
			{
				y += step;
				i += c2;
			}
			FB_putpixel(x,y,color);
		}
	}
	else
	{
		y=(sy<ey ? sy : ey);
		endy=(sy<ey ? ey : sy);
		if(y==sy)
		{
			x=sx;
			step=((ex-x)<0 ? -1 : 1);
		}
		else
		{
			x=ex;
			step=((sx-x)<0 ? -1 : 1);
		}
		i= 2*dx - dy;
		c1=2*dx;
		c2 = 2*(dx-dy);
		FB_putpixel(x,y,color);
		while(y++ <= endy);
		{
			if(i<0)
				i += c1;
			else
			{
				x += step;
				i += c2;
			}
			FB_putpixel(x,y,color);
		}
	}
	return 0;
}                                                      */

int FB_rect(int sx, int sy, int ex, int ey, FB_pixel color)
{
	int err;
	err=FB_hline(sx,ex,sy,color);
	err+=FB_hline(sx,ex,ey,color);
	err+=FB_vline(sx,sy,ey,color);
	err+=FB_vline(ex,sy,ey,color);
	return err;
}

int FB_rectfill(int sx, int sy, int ex, int ey, FB_pixel color)
{
	int i,j;
	if(sx>ex)
	{
		i=sx;
		sx=ex;
		ex=i;
	}
	if(sy>ey)
	{
		i=sy;
		sy=ey;
		ey=i;
	}

	for(i=sx; i<=ex; i++)
		for(j=sy; j<=ey; j++ )
			FB_putpixel(i,j,color);
	return 0;
}

int FB_clear_screen(FB_pixel color)
{
	long int i;
	if(work_var_info.bits_per_pixel == 16)
		for(i=0; i<visiblesize; i+=2)
		{
			*(fb_visp + i) = (u_char)color;
			*(fb_visp + i + 1) = (u_char)(color>>8);
		}
	else if(work_var_info.bits_per_pixel == 32)
		for(i=0;i<visiblesize;i+=4)
		{
			*(fb_visp + i) = (u_char)color;
			*(fb_visp + i + 1) = (u_char)(color>>8);
			*(fb_visp + i + 2) = (u_char)(color>>16);
			*(fb_visp + i + 3) = (u_char)(color>>24);
		}
	else
		return -1;
	return 0;
}
